require 'win32ole'

class CSSC
  def create_idl(svr, cls=svr)
    @itf_guid = WIN32OLE.create_guid.gsub(/[\{\}]/, "")
    @lib_guid = WIN32OLE.create_guid.gsub(/[\{\}]/, "")
    @cls_guid = WIN32OLE.create_guid.gsub(/[\{\}]/, "")
    str =<<EOF
[
    object,
    uuid(#{@itf_guid}),
    pointer_default(unique),
    dual
]
interface I#{cls}:IDispatch
{
    import "oaidl.idl";

};
[
    uuid(#{@lib_guid}),
    helpstring("#{svr} Type Library"),
    version(1.0)
]
library #{svr}lib
{
    importlib("stdole32.tlb");

    [
        uuid(#{@cls_guid}),
        helpstring("#{cls} class")
    ]
    coclass #{cls}
    {
        [default] interface I#{cls};
    };
};
EOF
  end

  def create_def(svr)
    src=<<EOF
LIBRARY      #{svr}
EXPORTS
    DllCanUnloadNow     private	
    DllGetClassObject   private	
    DllRegisterServer   private
    DllUnregisterServer private
EOF
  end

  def create_makefile(svr)
    f = svr.downcase
    src =<<EOF
#{f}.dll: iid_#{f}.obj #{f}.obj
\tlink /DLL /OUT:#{f}.dll /DEF:#{f}.def #{f}.obj iid_#{f}.obj user32.lib ole32.lib oleaut32.lib advapi32.lib
.c.obj:
\tcl -c $<
iid_#{f}.c:#{f}.idl
\tmidl /proxy nul /iid iid_#{f}.c /h #{f}.h /tlb #{f}.tlb #{f}.idl 
EOF
  end

  def create_cpp(svr, cls=svr)
    f = svr.downcase
    src =<<EOF
#include "#{f}.h"

class C#{cls}Factory : public IClassFactory
{
protected:
    ULONG    m_cRef;

public:
    C#{cls}Factory();
    ~C#{cls}Factory();

    //IUnknown...
    STDMETHODIMP        QueryInterface(REFIID, void**);
    STDMETHODIMP_(ULONG)    AddRef();
    STDMETHODIMP_(ULONG)    Release();

    //IClassFactory...
    STDMETHODIMP        CreateInstance(IUnknown*, REFIID, void**);
    STDMETHODIMP        LockServer(BOOL);

};

class C#{cls} : public I#{cls}
{
protected:
    ULONG        m_cRef;
public:
    C#{cls}();
    ~C#{cls}();

    //IUnknown...
    STDMETHODIMP        QueryInterface(REFIID, void**);
    STDMETHODIMP_(ULONG)    AddRef();
    STDMETHODIMP_(ULONG)    Release();

    //IDispatch...
    STDMETHODIMP        GetTypeInfoCount(UINT *);
    STDMETHODIMP        GetTypeInfo(UINT, LCID, ITypeInfo **);
    STDMETHODIMP        GetIDsOfNames(REFIID,OLECHAR**,UINT,LCID,DISPID *);
    STDMETHODIMP        Invoke(DISPID,REFIID,LCID,WORD,DISPPARAMS *,VARIANT *,EXCEPINFO *,UINT *);

};

UINT      g_cRefThisDll = 0;
HINSTANCE g_hmodThisDll = NULL;

BOOL SetRegKeyValue(LPTSTR pszKey, LPTSTR pszSubkey, LPTSTR pszName, LPTSTR pszValue)
{
    BOOL bOk = FALSE;
    LONG ec;
    HKEY hKey;
    TCHAR szKey[255];

    lstrcpy(szKey, pszKey);

    if (NULL != pszSubkey)
    {
        lstrcat(szKey, TEXT("\\\\"));
        lstrcat(szKey, pszSubkey);
    }

    ec = RegCreateKeyEx(HKEY_CLASSES_ROOT, szKey, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, NULL);

    if (ERROR_SUCCESS == ec) {
        if (NULL != pszValue) {
            ec = RegSetValueEx(hKey, pszName, 0, REG_SZ, (BYTE *)pszValue, (lstrlen(pszValue)+1)*sizeof(TCHAR));
        }
        if (ERROR_SUCCESS == ec)
            bOk = TRUE;
        RegCloseKey(hKey);
    }
    return bOk;
}

extern "C" int APIENTRY
DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
    if (dwReason == DLL_PROCESS_ATTACH)
    {
        g_hmodThisDll = hInstance;
        DisableThreadLibraryCalls(hInstance);
    }
    else if (dwReason == DLL_PROCESS_DETACH)
    {
    }
    return 1;
}

STDAPI DllCanUnloadNow(void)
{
    return (g_cRefThisDll == 0 ? S_OK : S_FALSE);
}

STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID *ppvOut)
{
    *ppvOut = NULL;

    if (IsEqualIID(rclsid, CLSID_#{cls}))
    {
        C#{cls}Factory *pcf = new C#{cls}Factory;

        return pcf->QueryInterface(riid, ppvOut);
    }

    return CLASS_E_CLASSNOTAVAILABLE;
}

STDAPI DllRegisterServer(void)
{
	TCHAR    szModulePath[MAX_PATH];

	GetModuleFileName(g_hmodThisDll, szModulePath, sizeof(szModulePath)/sizeof(TCHAR));
	SetRegKeyValue(TEXT("#{svr}Lib.#{cls}"), NULL, NULL, TEXT("#{svr}Lib"));
	SetRegKeyValue(TEXT("#{svr}Lib.#{cls}"), TEXT("CLSID"), NULL, TEXT("{#{@cls_guid}}"));
	SetRegKeyValue(TEXT("CLSID\\\\{#{@cls_guid}}"), NULL, NULL, TEXT("#{svr}Lib"));
	SetRegKeyValue(TEXT("CLSID\\\\{#{@cls_guid}}"), TEXT("InprocServer32"), NULL, szModulePath);
	SetRegKeyValue(TEXT("CLSID\\\\{#{@cls_guid}}\\\\InprocServer32"), NULL, TEXT("ThreadingModel"), TEXT("Apartment"));
	SetRegKeyValue(TEXT("CLSID\\\\{#{@cls_guid}}"), TEXT("ProgID"), NULL, TEXT("#{svr}Lib.#{cls}"));

	return S_OK;
}

STDAPI DllUnregisterServer(void)
{
	RegDeleteKey(HKEY_CLASSES_ROOT, TEXT("#{svr}Lib.#{cls}\\\\CLSID"));
	RegDeleteKey(HKEY_CLASSES_ROOT, TEXT("#{svr}Lib.#{cls}"));
	RegDeleteKey(HKEY_CLASSES_ROOT, TEXT("CLSID\\\\{#{@cls_guid}}\\\\InprocServer32"));
	RegDeleteKey(HKEY_CLASSES_ROOT, TEXT("CLSID\\\\{#{@cls_guid}}\\\\ProgID"));
	RegDeleteKey(HKEY_CLASSES_ROOT, TEXT("CLSID\\\\{#{@cls_guid}}"));

	return S_OK;
}

C#{cls}Factory::C#{cls}Factory()
{
    m_cRef = 0L;
    g_cRefThisDll++;	
}
																
C#{cls}Factory::~C#{cls}Factory()				
{
    g_cRefThisDll--;
}

STDMETHODIMP C#{cls}Factory::QueryInterface(REFIID riid, void** ppv)
{
    *ppv = NULL;
    if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IClassFactory))
    {
        *ppv = (IClassFactory*)this;
        AddRef();
        return S_OK;
    }
    return E_NOINTERFACE;
}	

STDMETHODIMP_(ULONG) C#{cls}Factory::AddRef()
{
    return ++m_cRef;
}

STDMETHODIMP_(ULONG) C#{cls}Factory::Release()
{
    if (--m_cRef)
        return m_cRef;
    delete this;
    return 0L;
}

STDMETHODIMP C#{cls}Factory::CreateInstance(IUnknown* pUnkOuter, REFIID riid, void **ppvObj)
{
    *ppvObj = NULL;
    if (pUnkOuter)
    	return CLASS_E_NOAGGREGATION;

    C#{cls} *pObj = new C#{cls}();

    if (NULL == pObj)
    	return E_OUTOFMEMORY;

    return pObj->QueryInterface(riid, ppvObj);
}

STDMETHODIMP C#{cls}Factory::LockServer(BOOL fLock)
{
    return S_OK;
}

C#{cls}::C#{cls}() {
	m_cRef = 0;
}

C#{cls}::~C#{cls}() {
}

STDMETHODIMP_(ULONG) C#{cls}::AddRef() {
    m_cRef++;
    return m_cRef;
}

STDMETHODIMP_(ULONG) C#{cls}::Release()
{
    if (--m_cRef)
        return m_cRef;
    delete this;
    return 0L;
}

STDMETHODIMP C#{cls}::QueryInterface(REFIID riid, void** ppv)
{
    *ppv = NULL;
    if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch))
    {
        *ppv = (IDispatch*)this;
        AddRef();
        return S_OK;
    }
    return E_NOINTERFACE;
}	

STDMETHODIMP C#{cls}::GetTypeInfoCount(UINT * pUint) {
    return E_NOTIMPL;
}

STDMETHODIMP C#{cls}::GetTypeInfo(UINT, LCID, ITypeInfo **) {
    return E_NOTIMPL;
}

STDMETHODIMP C#{cls}::GetIDsOfNames(
        const IID & iid, 
        LPOLESTR *pszNames,
        UINT cNames,
        LCID lcid,
        DISPID *pDispID) {
    for(UINT i = 0; i < cNames; i++) {
        pDispID[i] = DISPID_UNKNOWN;
        return DISP_E_MEMBERNOTFOUND;
    }
    return S_OK;
}

STDMETHODIMP C#{cls}::Invoke(
    DISPID id,
    const IID & iid,
    LCID lcid,
    WORD wFlags,
    DISPPARAMS *pDispParams,
    VARIANT *pVarResult,
    EXCEPINFO *pExcepInfo,
    UINT *pUArgError
    ) {
    int cArgs = pDispParams->cArgs;
    VARIANTARG *rgvarg = pDispParams->rgvarg;
    return DISPID_UNKNOWN;
}

EOF
  end

end

svr = ARGV.shift
sccs = CSSC.new
f = svr.downcase
idl_f = f + ".idl"
open(idl_f, "w") do |ofs|
  ofs.puts sccs.create_idl(svr)
end

cpp_f = f + ".cpp"
open(cpp_f, "w") do |ofs|
  ofs.puts sccs.create_cpp(svr)
end

def_f = f + ".def"
open(def_f, "w") do |ofs|
  ofs.puts sccs.create_def(svr)
end

make_f = "Makefile"
open(make_f, "w") do |ofs|
  ofs.puts sccs.create_makefile(svr)
end

