#include "mycomsvr.h"

class CMyComSvrFactory : public IClassFactory
{
protected:
    ULONG    m_cRef;

public:
    CMyComSvrFactory();
    ~CMyComSvrFactory();

    //IUnknown...
    STDMETHODIMP        QueryInterface(REFIID, void**);
    STDMETHODIMP_(ULONG)    AddRef();
    STDMETHODIMP_(ULONG)    Release();

    //IClassFactory...
    STDMETHODIMP        CreateInstance(IUnknown*, REFIID, void**);
    STDMETHODIMP        LockServer(BOOL);

};

class CMyComSvr : public IMyComSvr
{
protected:
    ULONG        m_cRef;
public:
    CMyComSvr();
    ~CMyComSvr();

    //IUnknown...
    STDMETHODIMP        QueryInterface(REFIID, void**);
    STDMETHODIMP_(ULONG)    AddRef();
    STDMETHODIMP_(ULONG)    Release();

    //IDispatch...
    STDMETHODIMP        GetTypeInfoCount(UINT *);
    STDMETHODIMP        GetTypeInfo(UINT, LCID, ITypeInfo **);
    STDMETHODIMP        GetIDsOfNames(REFIID,OLECHAR**,UINT,LCID,DISPID *);
    STDMETHODIMP        Invoke(DISPID,REFIID,LCID,WORD,DISPPARAMS *,VARIANT *,EXCEPINFO *,UINT *);

};

UINT      g_cRefThisDll = 0;
HINSTANCE g_hmodThisDll = NULL;

BOOL SetRegKeyValue(LPTSTR pszKey, LPTSTR pszSubkey, LPTSTR pszName, LPTSTR pszValue)
{
    BOOL bOk = FALSE;
    LONG ec;
    HKEY hKey;
    TCHAR szKey[255];

    lstrcpy(szKey, pszKey);

    if (NULL != pszSubkey)
    {
        lstrcat(szKey, TEXT("\\"));
        lstrcat(szKey, pszSubkey);
    }

    ec = RegCreateKeyEx(HKEY_CLASSES_ROOT, szKey, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, NULL);

    if (ERROR_SUCCESS == ec) {
        if (NULL != pszValue) {
            ec = RegSetValueEx(hKey, pszName, 0, REG_SZ, (BYTE *)pszValue, (lstrlen(pszValue)+1)*sizeof(TCHAR));
        }
        if (ERROR_SUCCESS == ec)
            bOk = TRUE;
        RegCloseKey(hKey);
    }
    return bOk;
}

extern "C" int APIENTRY
DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
    if (dwReason == DLL_PROCESS_ATTACH)
    {
        g_hmodThisDll = hInstance;
        DisableThreadLibraryCalls(hInstance);
    }
    else if (dwReason == DLL_PROCESS_DETACH)
    {
    }
    return 1;
}

STDAPI DllCanUnloadNow(void)
{
    return (g_cRefThisDll == 0 ? S_OK : S_FALSE);
}

STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID *ppvOut)
{
    *ppvOut = NULL;

    if (IsEqualIID(rclsid, CLSID_MyComSvr))
    {
        CMyComSvrFactory *pcf = new CMyComSvrFactory;

        return pcf->QueryInterface(riid, ppvOut);
    }

    return CLASS_E_CLASSNOTAVAILABLE;
}

STDAPI DllRegisterServer(void)
{
	TCHAR    szModulePath[MAX_PATH];

	GetModuleFileName(g_hmodThisDll, szModulePath, sizeof(szModulePath)/sizeof(TCHAR));
	SetRegKeyValue(TEXT("MyComSvrLib.MyComSvr"), NULL, NULL, TEXT("MyComSvrLib"));
	SetRegKeyValue(TEXT("MyComSvrLib.MyComSvr"), TEXT("CLSID"), NULL, TEXT("{FE8F8979-95D4-4BCD-AE94-FF14058C7407}"));
	SetRegKeyValue(TEXT("CLSID\\{FE8F8979-95D4-4BCD-AE94-FF14058C7407}"), NULL, NULL, TEXT("MyComSvrLib"));
	SetRegKeyValue(TEXT("CLSID\\{FE8F8979-95D4-4BCD-AE94-FF14058C7407}"), TEXT("InprocServer32"), NULL, szModulePath);
	SetRegKeyValue(TEXT("CLSID\\{FE8F8979-95D4-4BCD-AE94-FF14058C7407}\\InprocServer32"), NULL, TEXT("ThreadingModel"), TEXT("Apartment"));
	SetRegKeyValue(TEXT("CLSID\\{FE8F8979-95D4-4BCD-AE94-FF14058C7407}"), TEXT("ProgID"), NULL, TEXT("MyComSvrLib.MyComSvr"));

	return S_OK;
}

STDAPI DllUnregisterServer(void)
{
	RegDeleteKey(HKEY_CLASSES_ROOT, TEXT("MyComSvrLib.MyComSvr\\CLSID"));
	RegDeleteKey(HKEY_CLASSES_ROOT, TEXT("MyComSvrLib.MyComSvr"));
	RegDeleteKey(HKEY_CLASSES_ROOT, TEXT("CLSID\\{FE8F8979-95D4-4BCD-AE94-FF14058C7407}\\InprocServer32"));
	RegDeleteKey(HKEY_CLASSES_ROOT, TEXT("CLSID\\{FE8F8979-95D4-4BCD-AE94-FF14058C7407}\\ProgID"));
	RegDeleteKey(HKEY_CLASSES_ROOT, TEXT("CLSID\\{FE8F8979-95D4-4BCD-AE94-FF14058C7407}"));

	return S_OK;
}

CMyComSvrFactory::CMyComSvrFactory()
{
    m_cRef = 0L;
    g_cRefThisDll++;	
}
																
CMyComSvrFactory::~CMyComSvrFactory()				
{
    g_cRefThisDll--;
}

STDMETHODIMP CMyComSvrFactory::QueryInterface(REFIID riid, void** ppv)
{
    *ppv = NULL;
    if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IClassFactory))
    {
        *ppv = (IClassFactory*)this;
        AddRef();
        return S_OK;
    }
    return E_NOINTERFACE;
}	

STDMETHODIMP_(ULONG) CMyComSvrFactory::AddRef()
{
    return ++m_cRef;
}

STDMETHODIMP_(ULONG) CMyComSvrFactory::Release()
{
    if (--m_cRef)
        return m_cRef;
    delete this;
    return 0L;
}

STDMETHODIMP CMyComSvrFactory::CreateInstance(IUnknown* pUnkOuter, REFIID riid, void **ppvObj)
{
    *ppvObj = NULL;
    if (pUnkOuter)
    	return CLASS_E_NOAGGREGATION;

    CMyComSvr *pObj = new CMyComSvr();

    if (NULL == pObj)
    	return E_OUTOFMEMORY;

    return pObj->QueryInterface(riid, ppvObj);
}

STDMETHODIMP CMyComSvrFactory::LockServer(BOOL fLock)
{
    return S_OK;
}

CMyComSvr::CMyComSvr() {
	m_cRef = 0;
}

CMyComSvr::~CMyComSvr() {
}

STDMETHODIMP_(ULONG) CMyComSvr::AddRef() {
    m_cRef++;
    return m_cRef;
}

STDMETHODIMP_(ULONG) CMyComSvr::Release()
{
    if (--m_cRef)
        return m_cRef;
    delete this;
    return 0L;
}

STDMETHODIMP CMyComSvr::QueryInterface(REFIID riid, void** ppv)
{
    *ppv = NULL;
    if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch))
    {
        *ppv = (IDispatch*)this;
        AddRef();
        return S_OK;
    }
    return E_NOINTERFACE;
}	

STDMETHODIMP CMyComSvr::GetTypeInfoCount(UINT * pUint) {
    return E_NOTIMPL;
}

STDMETHODIMP CMyComSvr::GetTypeInfo(UINT, LCID, ITypeInfo **) {
    return E_NOTIMPL;
}

STDMETHODIMP CMyComSvr::GetIDsOfNames(
        const IID & iid, 
        LPOLESTR *pszNames,
        UINT cNames,
        LCID lcid,
        DISPID *pDispID) {
    for(UINT i = 0; i < cNames; i++) {
        pDispID[i] = DISPID_UNKNOWN;
        return DISP_E_MEMBERNOTFOUND;
    }
    return S_OK;
}

STDMETHODIMP CMyComSvr::Invoke(
    DISPID id,
    const IID & iid,
    LCID lcid,
    WORD wFlags,
    DISPPARAMS *pDispParams,
    VARIANT *pVarResult,
    EXCEPINFO *pExcepInfo,
    UINT *pUArgError
    ) {
    int cArgs = pDispParams->cArgs;
    VARIANTARG *rgvarg = pDispParams->rgvarg;
    return DISPID_UNKNOWN;
}

